/*
 * BPM calculator using a correlation algorithm variation
 * Copyright (C) 2012  Avelino Herrera Morales
 *
 * This program reads RAW signed mono 16bit/44100Hz audio files. To convert a
 * WAV file to this RAW format use the following sox command:
 *
 * $ sox input.wav -s -2 -c 1 -r 44100 output.raw
 *
 *
 * based on algorithm published at
 * http://archive.gamedev.net/archive/reference/programming/features/beatdetection/index.html
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <errno.h>
#include "fft.h"


#define  SAMPLE_RATE_HZ   44100
#define  BPM_PROBE_FROM   80
#define  BPM_PROBE_TO     160
#define  BPM_PROBE_STEP   5


#define  DEBUG  1


void generate_beat(uint16_t bpm, complex *output, uint32_t n) {
	uint32_t i;
	uint32_t distance = (60 * SAMPLE_RATE_HZ) / ((uint32_t) bpm);
	for (i = 0; i < n; i++) {
		output[i].re = ((i % distance) == 0) ? 1 : 0;
		output[i].im = 0;
	}
}


void fill_fft_complex_from_float(float *source, uint32_t source_size, complex **dest, uint32_t *log2_dest_size) {
	uint32_t i;
	uint32_t dest_size;
	*log2_dest_size = (uint32_t) ceil(log(source_size) / log(2));
	dest_size = 1 << (*log2_dest_size);
	*dest = (complex *) malloc(dest_size * sizeof(complex));
	if (*dest == NULL)
		return;
	for (i = 0; i < dest_size; i++) {
		if (i < source_size)
			(*dest)[i].re = source[i];
		else
			(*dest)[i].re = 0;
		(*dest)[i].im = 0;
	}
}


uint16_t bpm_detect(float *music_samples, uint32_t n, int bpm_from, int bpm_to, int bpm_step) {
	uint16_t bpm;
	float max_energy;
	uint16_t max_energy_bpm;
	complex *complex_music_samples;
	complex *complex_beat_samples;
	uint32_t log2size;
	uint32_t size;
	uint32_t i;
	fill_fft_complex_from_float(music_samples, n, &complex_music_samples, &log2size);
	if (complex_music_samples == NULL)
		return;
	size = 1 << log2size;
	fft(complex_music_samples, log2size);
	if ((complex_beat_samples = (complex *) malloc(size * sizeof(complex))) == NULL) {
		free(complex_music_samples);
		return;
	}
#ifdef  DEBUG
	printf("original size = %d\n", (int) n);
	printf("new size = %d\n", (int) size);
	printf("log2 new size = %d\n", (int) log2size);
#endif
	max_energy = 0;
	for (bpm = bpm_from; bpm <= bpm_to; bpm += bpm_step) {
		float energy;
		generate_beat(bpm, complex_beat_samples, size);
		fft(complex_beat_samples, log2size);
		energy = 0;
		for (i = 0; i < size; i++) {
			float aux = fabs((complex_music_samples[i].re * complex_beat_samples[i].re) + (complex_music_samples[i].im * complex_beat_samples[i].im));
			energy += aux;
		}
#ifdef  DEBUG
		printf("energy for %3d is %f\n", (int) bpm, energy);
#endif
		if (energy > max_energy) {
			max_energy = energy;
			max_energy_bpm = bpm;
		}
	}
	free(complex_beat_samples);
	free(complex_music_samples);
	return max_energy_bpm;
}


void read_raw_file(char *path, float **dest, uint32_t *n) {
	FILE *f;
	long file_size;
	int16_t *raw_data;
	uint32_t i;
	if ((f = fopen(path, "r")) == NULL) {
		perror("fopen");
		*n = 0;
		return;
	}
	if (fseek(f, 0, SEEK_END) != 0) {
		perror("fseek");
		*n = 0;
		fclose(f);
		return;
	}
	if ((file_size = ftell(f)) < 0) {
		perror("ftell");
		*n = 0;
		fclose(f);
		return;
	}
	if (fseek(f, 0, SEEK_SET) != 0) {
		perror("fseek");
		*n = 0;
		fclose(f);
		return;
	}
	*n = file_size >> 1;
	raw_data = (int16_t *) malloc(*n * sizeof(int16_t));
	if (raw_data == NULL) {
		perror("malloc");
		*n = 0;
		fclose(f);
		return;
	}
	fread(raw_data, 1, file_size, f);
	fclose(f);
	*dest = (float *) malloc(*n * sizeof(float));
	if (*dest == NULL) {
		perror("malloc");
		free(raw_data);
		*n = 0;
		return;
	}
	for (i = 0; i < *n; i++)
		(*dest)[i] = ((float) raw_data[i]) / 32768;
	free(raw_data);
}


int main(int argc, char *argv[]) {
	float *music_samples;
	uint32_t n;
	uint16_t bpm;
	if (argc < 2) {
		printf("syntax: %s path_to_raw_file\n", argv[0]);
		return 1;
	}
	read_raw_file(argv[1], &music_samples, &n);
	bpm = bpm_detect(music_samples, n, BPM_PROBE_FROM, BPM_PROBE_TO, BPM_PROBE_STEP);
	printf("bpm = %d\n", (int) bpm);
	return 0;
}
